///////////////////////////////////////////////////////////////////////////////////////
// 
// DCM AHRS Filter
///////////////////////////////////////////////////////////////////////////////////////

/***************************************************************************************************************
* Ra_zor AHRS Firmware v1.4.2
* 9 Degree of Measurement Attitude and Heading Reference System
* for Sparkfun "9DOF Ra_zor IMU" (SEN-10125 and SEN-10736)
* and "9DOF Sensor Stick" (SEN-10183, 10321 and SEN-10724)
*
* Released under GNU GPL (General Public License) v3.0
* Copyright (C) 2013 Peter Bartz [http://ptrbrtz.net]
* Copyright (C) 2011-2012 Quality & Usability Lab, Deutsche Telekom Laboratories, TU Berlin
*
* Infos, updates, bug reports, contributions and feedback:
*     https://github.com/ptrbrtz/ra_zor-9dof-ahrs
*
*
* History:
*   * Original code (http://code.google.com/p/sf9domahrs/) by Doug Weibel and Jose Julio,
*     based on ArduIMU v1.5 by Jordi Munoz and William Premerlani, Jose Julio 
*/

/*
  a_xis definition (differs from definition printed on the board!):
    X a_xis pointing forward (towards the short edge with the connector holes)
    Y a_xis pointing to the right
    and Z a_xis pointing down.
    
  Positive yaw   : clockwise
  Positive roll  : right wing down
  Positive pitch : nose up
  
  Transformation order: first yaw then pitch then roll.
*/


#include "DCM.h"
#include <Arduino.h>

DCM::DCM(){

}

void DCM::DCM_init(float Kp_RP, float Ki_RP, float Kp_Y, float Ki_Y)
{

  //
  //Get Proportional and Integral Gains set in FreeIMU Library
  // 
  Kp_ROLLPITCH = Kp_RP;
  Ki_ROLLPITCH = Ki_RP;
  Kp_YAW = Kp_Y;
  Ki_YAW = Ki_Y;

  void reset_sensor_fusion();
}

//------------------- MATH.INO --------------------

/* This file is part of the Ra_zor AHRS Firmware */

// Computes the dot product of two vectors
float DCM::Vector_Dot_Product(float *v1, float *v2)
{
  float result = 0;
  
  for(int c = 0; c < 3; c++)
  {
    result += v1[c] * v2[c];
  }
  
  return result; 
}

// Computes the cross product of two vectors
// out has to different from v1 and v2 (no in-place)!
void DCM::Vector_Cross_Product(float *out,  float *v1,  float *v2)
{
  out[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
  out[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
  out[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
}

// Multiply the vector by a scalar
void DCM::Vector_Scale(float out[3],  float v[3], float scale)
{
  for(int c = 0; c < 3; c++)
  {
    out[c] = v[c] * scale; 
  }
}

// Adds two vectors
void DCM::Vector_Add(float out[3],  float v1[3],  float v2[3])
{
  for(int c = 0; c < 3; c++)
  {
    out[c] = v1[c] + v2[c];
  }
}

// Multiply two 3x3 matrices: out = a * b
// out has to different from a and b (no in-place)!
void DCM::Matrix_Multiply( float a[3][3], float b[3][3], float out[3][3])
{
  for(int x = 0; x < 3; x++)  // rows
  {
    for(int y = 0; y < 3; y++)  // columns
    {
      out[x][y] = a[x][0] * b[0][y] + a[x][1] * b[1][y] + a[x][2] * b[2][y];
    }
  }
}

// Multiply 3x3 matrix with vector: out = a * b
// out has to different from b (no in-place)!
void DCM::Matrix_Vector_Multiply( float a[3][3], float b[3], float out[3])
{
  for(int x = 0; x < 3; x++)
  {
    out[x] = a[x][0] * b[0] + a[x][1] * b[1] + a[x][2] * b[2];
  }
}

// Init rotation matrix using euler angles
void DCM::init_rotation_matrix(float m[3][3], float yaw, float pitch, float roll)
{
  float c1 = cos(roll);
  float s1 = sin(roll);
  float c2 = cos(pitch);
  float s2 = sin(pitch);
  float c3 = cos(yaw);
  float s3 = sin(yaw);

  // Euler angles, right-handed, intrinsic, XYZ convention
  // (which means: rotate around body a_xes Z, Y', X'') 
  m[0][0] = c2 * c3;
  m[0][1] = c3 * s1 * s2 - c1 * s3;
  m[0][2] = s1 * s3 + c1 * c3 * s2;

  m[1][0] = c2 * s3;
  m[1][1] = c1 * c3 + s1 * s2 * s3;
  m[1][2] = c1 * s2 * s3 - c3 * s1;

  m[2][0] = -s2;
  m[2][1] = c2 * s1;
  m[2][2] = c1 * c2;
}

// Read every sensor and record a time stamp
// Init DCM with unfiltered orientation
// TODO re-init global vars?
void DCM::reset_sensor_fusion() {
  float temp1[3];
  float temp2[3];
  float xa_xis[] = {1.0f, 0.0f, 0.0f};

  //read_sensors();
  timestamp = millis();
  
  // GET PITCH
  // Using y-z-plane-component/x-component of gravity vector
  pitch = -atan2(accel[0], sqrt(accel[1] * accel[1] + accel[2] * accel[2]));
	
  // GET ROLL
  // Compensate pitch of gravity vector 
  Vector_Cross_Product(temp1, accel, xa_xis);
  Vector_Cross_Product(temp2, xa_xis, temp1);
  
  // Normally using x-z-plane-component/y-component of compensated gravity vector
  // roll = atan2(temp2[1], sqrt(temp2[0] * temp2[0] + temp2[2] * temp2[2]));
  // Since we compensated for pitch, x-z-plane-component equals z-component:
  roll = atan2(temp2[1], temp2[2]);

  // GET YAW
  //Compass_Heading();  //NEED TO SUBSTITUTE ICOMPASS
  yaw = MAG_Heading;  
    
  // Init rotation matrix
  init_rotation_matrix(DCM_Matrix, yaw, pitch, roll);

}  

//------------- DCM.INO -------------------------

/* This file is part of the Ra_zor AHRS Firmware */

// DCM algorithm

/**************************************************/
void DCM::Normalize(void)
{
  float error=0;
  float temporary[3][3];
  float renorm=0;
  
  error= -Vector_Dot_Product(&DCM_Matrix[0][0],&DCM_Matrix[1][0])*.5; //eq.19

  Vector_Scale(&temporary[0][0], &DCM_Matrix[1][0], error); //eq.19
  Vector_Scale(&temporary[1][0], &DCM_Matrix[0][0], error); //eq.19
  
  Vector_Add(&temporary[0][0], &temporary[0][0], &DCM_Matrix[0][0]);//eq.19
  Vector_Add(&temporary[1][0], &temporary[1][0], &DCM_Matrix[1][0]);//eq.19
  
  Vector_Cross_Product(&temporary[2][0],&temporary[0][0],&temporary[1][0]); // c= a x b //eq.20
  
  renorm= .5 *(3 - Vector_Dot_Product(&temporary[0][0],&temporary[0][0])); //eq.21
  Vector_Scale(&DCM_Matrix[0][0], &temporary[0][0], renorm);
  
  renorm= .5 *(3 - Vector_Dot_Product(&temporary[1][0],&temporary[1][0])); //eq.21
  Vector_Scale(&DCM_Matrix[1][0], &temporary[1][0], renorm);
  
  renorm= .5 *(3 - Vector_Dot_Product(&temporary[2][0],&temporary[2][0])); //eq.21
  Vector_Scale(&DCM_Matrix[2][0], &temporary[2][0], renorm);
}

/**************************************************/
void DCM::Drift_correction(void)
{
  float mag_heading_x;
  float mag_heading_y;
  float errorCourse;

  //Compensation the Roll, Pitch and Yaw drift. 
  static float Scaled_Omega_P[3];
  static float Scaled_Omega_I[3];
  float Accel_magnitude;
  float Accel_weight;
  
  //*****Roll and Pitch***************
  // Calculate the magnitude of the accelerometer vector

  Accel_magnitude = sqrt(Accel_Vector[0]*Accel_Vector[0] + Accel_Vector[1]*Accel_Vector[1] + Accel_Vector[2]*Accel_Vector[2]);
  Accel_magnitude = Accel_magnitude / GRAVITY; // Scale to gravity.

  // Dynamic weighting of accelerometer info (reliability filter)
  // Weight for accelerometer info (<0.5G = 0.0, 1G = 1.0 , >1.5G = 0.0)
  Accel_weight = constrain(1 - 2*abs(1 - Accel_magnitude),0,1);  //  

  Vector_Cross_Product(&errorRollPitch[0],&Accel_Vector[0],&DCM_Matrix[2][0]); //adjust the ground of reference
  Vector_Scale(&Omega_P[0],&errorRollPitch[0],Kp_ROLLPITCH*Accel_weight);
  
  Vector_Scale(&Scaled_Omega_I[0],&errorRollPitch[0],Ki_ROLLPITCH*Accel_weight);
  Vector_Add(Omega_I,Omega_I,Scaled_Omega_I);     
  
  //*****YAW***************
  // We make the gyro YAW drift correction based on compass magnetic heading
  
  mag_heading_x = cos(MAG_Heading);
  mag_heading_y = sin(MAG_Heading);
  errorCourse=(DCM_Matrix[0][0]*mag_heading_y) - (DCM_Matrix[1][0]*mag_heading_x);  //Calculating YAW error
  Vector_Scale(errorYaw,&DCM_Matrix[2][0],errorCourse); //Applys the yaw correction to the XYZ rotation of the aircraft, depeding the position.
  
  Vector_Scale(&Scaled_Omega_P[0],&errorYaw[0],Kp_YAW);//.01proportional of YAW.
  Vector_Add(Omega_P,Omega_P,Scaled_Omega_P);//Adding  Proportional.
  
  Vector_Scale(&Scaled_Omega_I[0],&errorYaw[0],Ki_YAW);//.00001Integrator
  Vector_Add(Omega_I,Omega_I,Scaled_Omega_I);//adding integrator to the Omega_I
}


void DCM::Matrix_update(void)
{

  gyro_Vector[0] = TO_RAD(gyro[0]); //gyro x roll 
  gyro_Vector[1] = TO_RAD(gyro[1]); //gyro y pitch
  gyro_Vector[2] = TO_RAD(gyro[2]); //gyro z yaw 

  Accel_Vector[0]=accel[0];
  Accel_Vector[1]=accel[1];
  Accel_Vector[2]=accel[2];

  Vector_Add(&Omega[0], &gyro_Vector[0], &Omega_I[0]);  //adding proportional term
  Vector_Add(&Omega_Vector[0], &Omega[0], &Omega_P[0]); //adding Integrator term
  
#if DEBUG__NO_DRIFT_CORRECTION == true // Do not use drift correction
  Update_Matrix[0][0]=0;
  Update_Matrix[0][1]=-G_Dt*gyro_Vector[2];//-z
  Update_Matrix[0][2]=G_Dt*gyro_Vector[1];//y
  Update_Matrix[1][0]=G_Dt*gyro_Vector[2];//z
  Update_Matrix[1][1]=0;
  Update_Matrix[1][2]=-G_Dt*gyro_Vector[0];
  Update_Matrix[2][0]=-G_Dt*gyro_Vector[1];
  Update_Matrix[2][1]=G_Dt*gyro_Vector[0];
  Update_Matrix[2][2]=0;
#else // Use drift correction
  Update_Matrix[0][0]=0;
  Update_Matrix[0][1]=-G_Dt*Omega_Vector[2];//-z
  Update_Matrix[0][2]=G_Dt*Omega_Vector[1];//y
  Update_Matrix[1][0]=G_Dt*Omega_Vector[2];//z
  Update_Matrix[1][1]=0;
  Update_Matrix[1][2]=-G_Dt*Omega_Vector[0];//-x
  Update_Matrix[2][0]=-G_Dt*Omega_Vector[1];//-y
  Update_Matrix[2][1]=G_Dt*Omega_Vector[0];//x
  Update_Matrix[2][2]=0;
#endif
	
 Matrix_Multiply(DCM_Matrix, Update_Matrix, Temporary_Matrix); //a*b=c

  for(int x=0; x<3; x++) //Matrix Addition (update)
  {
    for(int y=0; y<3; y++)
    {
      DCM_Matrix[x][y] += Temporary_Matrix[x][y];
    }
  }
}

void DCM::getEulerRad(float * angles)
{
  angles[1] = asin(DCM_Matrix[2][0]);  //originally -asin, pitch
  angles[2] = atan2(DCM_Matrix[2][1],DCM_Matrix[2][2]);  //roll
  angles[0] = -atan2(DCM_Matrix[1][0],DCM_Matrix[0][0]);  //yaw
}

void DCM::getEulerDeg(float * angles)
{
  angles[1] = TO_DEG(-asin(DCM_Matrix[2][0]));
  angles[2] = TO_DEG(atan2(DCM_Matrix[2][1],DCM_Matrix[2][2]));
  angles[0] = TO_DEG(atan2(DCM_Matrix[1][0],DCM_Matrix[0][0]));
}

//
//Set sensor values received from FreeIMU library
//
void DCM::setSensorVals(float g_x, float g_y, float g_z, float a_x, float a_y, float a_z, float m_x, float m_y, float m_z, float magHead)
{
  float normA = sqrt(a_x*a_x + a_y*a_y + a_z*a_z);
  accel[0] = a_x/normA;   accel[1] = a_y/normA;  accel[2] = a_z/normA;
  gyro[0] = g_x;   gyro[1] = g_y;  gyro[2] = g_z;
  float normM = sqrt(m_x*m_x + m_y*m_y + m_z*m_z);
  magnetom[0] = m_x/normM; magnetom[1] = m_y/normM; magnetom[2] = m_z/normM;
	MAG_Heading = magHead;
}


//
//DCM to Quaternion conversion based on Vector Nav Application Note
//AN002, Quaternion Math
// 
void DCM::getDCM2Q(float * qd)
{
  float angles[3];
  //quaternions as defined in Vector Nav App Note AN002
  //q3 = 0.5 * sqrt(DCM_Matrix[0][0]+DCM_Matrix[1][1]+DCM_Matrix[2][2]+1);
  //q0 = (DCM_Matrix[1][2]-DCM_Matrix[2][1])/(4*q3);
  //q1 = (DCM_Matrix[2][0]-DCM_Matrix[0][2])/(4*q3);
  //q2 = (DCM_Matrix[0][1]-DCM_Matrix[1][0])/(4*q3);
  //convert to quaternion notation as used in the FreeIMU library
  //qd[0] =  q3;
  //qd[1] = -q0;
  //qd[2] = -q1;
  //qd[3] = -q2;
  getEulerRad(angles);
  
  float cy = cos(angles[0] * 0.5);
  float sy = sin(angles[0] * 0.5);
  float cr = cos(angles[2] * 0.5);
  float sr = sin(angles[2] * 0.5);
  float cp = cos(angles[1] * 0.5);
  float sp = sin(angles[1] * 0.5);

  qd[0] = cy * cr * cp + sy * sr * sp;
  qd[1] = cy * sr * cp - sy * cr * sp;
  qd[2] = cy * cr * sp + sy * sr * cp;
  qd[3] = sy * cr * cp - cy * sr * sp;
}

void DCM::calDCM() 
{
    Matrix_update();
    Normalize();
    Drift_correction();
}



