///////////////////////////////////////////////////////////////////////////////////////
// 
// Library for Popular AHRS Algorithms
///////////////////////////////////////////////////////////////////////////////////////

#include "MPU9250.h"
#include "EEPROM.h"
#include "globals.h"
#include "UBLOX.h"
#include "A_ConfigDefines.h"

#include "Eigen.h"
#include <Eigen/Dense>

//---- Filter Selection ------
// Set filter type: 
//  0 = Madgwick implementation of Mahoney DCM
//  1 = Madgwick Gradient Descent, in Quaternion form
//  2 = Mahony, in Quaternion form
//  3 = Madwick Original Paper AHRS 
//  4 = DCM Implementation
//  5 = Complimentary Filter
//  6 = uNavEulAHRS

// Select the filter type here
#define MARG 3

uint32_t GPSrxs=0; // bugbug debug

// Uncomment this line to disable the magnetometer in the 
// sensor fusion algorithm
//#define DISABLE_MAGN 

#if MARG == 0
  #include "AHRS.h"
#elif MARG == 1
  #include "MadgwickAHRS.h"
#elif MARG == 2
  #include "MahonyAHRS.h"
#elif MARG == 3
  #include "MargUpdateFilter.h"
#elif MARG == 4
  #include "DCM.h"
  DCM dcm;
#elif MARG == 5
  #include "compFlt.h"
#elif MARG == 6
  #include "uNavEulAHRS.h"
  // a uNavAHRS object
  uNavAHRS Filter;
#endif

// an MPU-9250 object on SPI bus 0 with chip select 10
MPU9250 Imu(IMU_BUS, IMU_ADDR);
int status;

UBLOX gps(GPS_Start);
bool newGPSData;
unsigned long prevTOW;
gpsData uBloxData;
//unsigned long tstart, tstop;

void setup() {
  // serial to display data
  cout.begin(cout_BAUD);
  #ifdef coutD
    coutD.begin(coutD_BAUD);
  #endif

  while(!cout && millis() < 2000 ) {}
  delay(5000);

  gps.begin(GPS_BAUD);
  if ( 255 != GPS_AltRx ) GPS_PORT.setRX(GPS_AltRx);
  attachInterrupt(digitalPinToInterrupt(GPS_SRX), GPS_serialrx_isr, RISING);

  //init imu
  // start communication with IMU 
  if ( 255 != IMU_SCL ) Wire.setSCL(IMU_SCL);
  if ( 255 != IMU_SDA ) Wire.setSDA(IMU_SDA);
  status = Imu.begin();
  if ( 0 != IMU_SPD ) Wire.setClock(IMU_SPD);
   
  initIMU();

  // quaternion of sensor frame relative to auxiliary frame
  q[0] = 1.0f; q[1] = 0.0f; 
  q[2] = 0.0f; q[3] = 0.0f;  

  // Time parameter initialization
  sampleTime = 0;
  _dt = 0.001;
  tsGPS = 0;
  tsIMU = 0;
}

void loop(){

  readGPS();
      // Calculate utcTime for this loop iteration
      utcTime = float(uBloxData.utcHour)*60.0f*60.0f +
                float(uBloxData.utcMin) *60.0f +
                float(uBloxData.utcSec) +
                float(uBloxData.utcNano)/1000000000.0;
  if(rtk.upDated){
          timeStamp = 0.0f;
          serialDataOut();
  } else {
    utcTime = 0.0f;
  }
  //gps.read(&uBloxData);

  // Wait for new data from IMU to run filter algorithms
  if (newData == 1) {
    newData = 0;
    sampleTime = 0;
    
    // read the sensor
    Imu.readSensor();
    
    // update the filter
    gx1 = Imu.getGyroX_rads(); gy1 = Imu.getGyroY_rads(); gz1 = Imu.getGyroZ_rads();
    ax1 = Imu.getAccelX_mss(); ay1 = Imu.getAccelY_mss(); az1 = Imu.getAccelZ_mss();
    mx1 = Imu.getMagX_uT();    my1 = Imu.getMagY_uT();    mz1 = Imu.getMagZ_uT();  
    //if (Filter.update(gx,gy,gz,ax,ay,az,mx,my,mz)) {
    //  tstop = micros();

    now = micros();
    #if MARG != 6
      _dt = ((now - lastUpdate) * 0.000001);
    #endif
     
    #if(MARG == 0 || MARG == 1 || MARG == 2 || MARG == 3)  
      #if (not defined(DISABLE_MAGN))
        #if MARG == 0
            AHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #elif MARG == 1
            MadgwickAHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);        
        #elif MARG == 2
            MahonyAHRSupdate(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #else
            MARGUpdateFilter(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
        #endif
      #else
        #if MARG == 0
            AHRSupdateIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #elif MARG == 1
            MadgwickAHRSupdateIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #else
            MARGUpdateFilterIMU(gx1, gy1, gz1, ax1, ay1, az1);
        #endif
      #endif
    #elif (MARG == 4)
        //val[9] = maghead.iheading(1, 0, 0, val[0], val[1], val[2], val[6], val[7], val[8]);
        magHead = calcMagHeading(q[0], q[1], q[2], q[3], mx1, my1, mz1);
        dcm.setSensorVals(gx1*180./PI, gy1*180./PI, gz1*180./PI, ax1, ay1, az1, mx1, my1, mz1, magHead);
        dcm.G_Dt = _dt;
        dcm.calDCM();
        dcm.getDCM2Q(q);
    #elif (MARG == 5)
        //compFilter(gx1, gy1, gz1, ax1, ay1, az1, 0, 0, 0);    //yaw based on filter alone
        compFilter(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);//yaw calculated with magnetometer
    #elif MARG == 6
        upDated = Filter.update(gx1, gy1, gz1, ax1, ay1, az1, mx1, my1, mz1);
    #endif

    lastUpdate = now;
    if(sincePrint > 25){
      sincePrint = 0;
      #if (MARG == 5)
        //do nothing
      #elif(MARG == 4)
        dcm.getEulerDeg(ypr);
      #elif (MARG == 6)
        ypr[0] = Filter.getYaw_rad()*180.0f/PI;
        ypr[1] = Filter.getPitch_rad()*180.0f/PI;
        ypr[2] = Filter.getRoll_rad()*180.0f/PI;
        //cout.println(Filter.getHeading_rad()*180.0f/PI);
      #else
        getYawPitchRoll(ypr);
      #endif

      // Rotate accel body to accel inertial (NED)
      body2inertial();

      // Send data to serial port. Check for duplicate GPS readings first.
      //if(sincePrint > 25){
          serialDataOut();
      }
  }

  //Serial passthrough so you can see GPS data
  //via uCenter.  Doesn't read/write configs
  if (cout.available() > 0)
  {
    int val = cout.read();  //read telem input commands        
    switch(val)
    {
      case 'p':
        toggleGPS();
        break;
      default:
        break;
    }
  }
}

// This routine is called via interrupt from the IMU
void runFilter() {
  newData = 1;
  //IMU_RX_time = tsIMU;
}


//========================================
#define RxGapTime ((F_CPU / GPS_BAUD) * 30)
//==================================
// FASTRUN void GPS_serialrx_isr()
// GLOBALS :: volatile uint32_t GPS_RX_Time = 0;
FASTRUN void GPS_serialrx_isr() {
    GPS_RX_time = tsGPS; // Start bit transition detected after 'idle' time
    detachInterrupt(digitalPinToInterrupt(GPS_SRX));
    GPS_newData=1;
    GPSrxs++;
}
//==================================
void toggleGPS(){
  if(passThrough_toggle == 0){
   passThrough_toggle = 1;
   passThroughMode();
  } else {
    passThrough_toggle = 0;
  }
}
